/*
   Process the Makefile.in files.
 */

/* Copyright, compile requirements.

   (c) Keith Owens <kaos@ocs.com.au> February 2002.
   Released under GPL version 2, (you know the rest).
 */

/* Run requirements.

   The environment must contain the correct variables, see pp_getenv.c.
 */

/* Processing.
 
   Using the kbuild database, read all the Makefile.in files and convert them
   from the kbuild mini language to make rules.  Update the database as new
   targets are identified.

   Outputs are all in the object tree.

     .tmp_db_main.  The database used by all pre-processor code.

     .tmp_global_makefile_in.1.  All the Makefile.in files concatenated in
     the required order.

     .tmp_global_makefile_in.2.  .tmp_global_makefile_in.1 converted from
     kbuild commands to make rules.
 */

/* Parameters.

   --verbose		Display what pp_makefile2 is doing.
 */

#include "pp_makefile.h"
#include "pp_db.h"

#include <dirent.h>
#include PP_GETOPT_H
#include <limits.h>
#include <signal.h>

const char *program = "pp_makefile2";

/* Defined in the parser */
extern int parse_makefile_in(PPDB * db);

int errors;
int norun;			/* Required by pp_db but irrelevant for this program */
int verbose;
int signalled;

/*============================================================================*/

/* Copy an individual Makefile.in fragment to the output file. */

static int
copy_one_makefile_in(PPDB * db, DB_REC_CHILD_1 * c1, void *userdata)
{
	DB_REC_DIRECTORY *d;
	DB_TOKEN token;
	FILE *in, *f = userdata;
	char makefile[PATH_MAX + 1], dirname[PATH_MAX + 1], *line;
	const char *relname;
	int eof, line_size, ret = 0, relname_len;
	if (!c1->flags.primary || c1->entry.type != DB_RECTYPE_SOURCE) {
		return (0);
	}
	relname = read_db_text(db, c1->relname);
	relname_len = strlen(relname);
	if (relname_len < 12
	    || strncmp(relname + relname_len - 12, "/Makefile.in", 12)) {
		return (0);
	}
	/* Ignore Makefile.in from arch directories other than the current arch */
	if (relname_len > 6 && strncmp(relname, "/arch/", 6) == 0) {
		if (strncmp(relname + 6, arch, arch_len) ||
		    *(relname + 6 + arch_len) != '/') {
			VERBOSE(2, "   ignoring %s\n", relname);
			return (0);
		}
	}
	VERBOSE(2, "   copying %s\n", relname);
	snprintf(dirname, sizeof (dirname), "%.*s", relname_len - 11, relname);
	token = lookup_db_directory(db, dirname);
	d = read_db_directory(db, token);
	d->flags.makefile = 1;
	fprintf(f, "#pp_makefile_in_start %s %s\n", srctree[c1->index],
		dirname);
	snprintf(makefile, sizeof (makefile), "%s%s", srctree[c1->index],
		 relname + 1);
	if (!(in = fopen(makefile, "r"))) {
		fprintf(stderr, "%s: Cannot open %s for reading: %m\n",
			program, makefile);
		ret = 1;
	}
	if (!ret) {
		line = k_malloc((line_size = 4096));
		eof = 0;
		while (!(eof = fgets_long_line(&line, &line_size, in))) {
			fprintf(f, "%s", line);
		}
		if (eof < 0) {
			fprintf(stderr,
				"%s: Error reading from %s: %m\n",
				program, makefile);
			ret = 1;
		}
		free(line);
		fclose(in);
		fprintf(f, "\n#pp_makefile_in_end\n");
	}
	return (ret);
}

/* Copy all the Makefile.in fragments to the output file */

static int
copy_makefile_in(PPDB * db)
{
	FILE *global_makefile;
	VERBOSE(1, "   copy_makefile_in\n");
	if (!(global_makefile = fopen(".tmp_global_makefile_in.1", "w"))) {
		fprintf(stderr,
			"%s: Cannot open .tmp_global_makefile_in.1 for writing: %m\n",
			program);
		return (1);
	}
	run_db_child(db, copy_one_makefile_in, global_makefile, "/");
	fclose(global_makefile);
	return (0);
}

/*============================================================================*/

/* Clear old data from the database.  Empty most target and directory flags  */

static int
clear_old_data_run(PPDB * db, kvpair kv, void *userdata)
{
	DB_TOKEN token;
	memcpy(&token, (void *) (kv.key.dptr), sizeof (token));
	if (token.type == DB_RECTYPE_TARGET) {
		DB_REC_TARGET *t = (DB_REC_TARGET *) (kv.val.dptr), temp;
		temp.flags = t->flags;
		memset(&(t->flags), 0, sizeof (t->flags));
		t->flags.uptodate = temp.flags.uptodate;
		t->flags.ignore = temp.flags.ignore;
		t->flags.symlink = temp.flags.symlink;
	}
	if (token.type == DB_RECTYPE_DIRECTORY) {
		DB_REC_DIRECTORY *d = (DB_REC_DIRECTORY *) (kv.val.dptr), temp;
		temp.flags = d->flags;
		memset(&(d->flags), 0, sizeof (d->flags));
		d->flags.explicit = temp.flags.explicit;
	}
	return (0);
}

static int
clear_old_data(PPDB * db)
{
	return (run_db(db, clear_old_data_run, NULL));
}

/*============================================================================*/

static void
usage(void)
{
	fprintf(stderr, "usage: %s [--verbose]\n", program);
	exit(1);
}

static int
parse(int argc, char **argv)
{
	int o;
	static struct option long_opts[] = {
		{"verbose", 0, 0, 'v'},
		{0, 0, 0, 0}
	};

	/* Parse the command line.  */
	while ((o = getopt_long(argc, argv, "v", &long_opts[0], NULL)) != EOF) {
		switch (o) {
		case 'v':	/* --verbose */
			++verbose;
			break;
		default:
			usage();
			break;
		}
	}

	if (optind != argc) {
		usage();
	}

	return (errors);
}

static void
handler(int signum)
{
	if (++signalled > 2) {
		exit(2);
	}
}

int
main(int argc, char **argv)
{
	PPDB *db_main;
	int ret = 0;
	signal(SIGHUP, handler);
	signal(SIGINT, handler);
	signal(SIGQUIT, handler);
	signal(SIGPIPE, handler);
	signal(SIGTERM, handler);
	if (pp_getenv() || parse(argc, argv)) {
		return (1);
	}
	VERBOSE(1, "%s verbose %d\n", program, verbose);
	db_main = open_db_main(1);
	if (!db_main || signalled ||
	    clear_old_data(db_main) || signalled ||
	    copy_makefile_in(db_main) || signalled ||
	    parse_makefile_in(db_main) || signalled) {
		ret = 1;
	}
	if (db_main) {
		close_db_main(db_main);
	}
	pp_freeenv();
	return (errors || ret);
}
